#!/usr/bin/env python3
r"""
Used by CI to bootstrap a new cluster and run tests.

Env vars needed:
- COSMOS_API_TOKEN

Install requirements:

- Alpine using `requirements.txt`:

    apk --no-cache add python3-dev build-base libffi-dev linux-headers \
      openssl-dev openssh-client
    pip3 install -r requirements.txt

- Apline using packages (much faster):

    apk --no-cache add ansible musl-dev linux-headers gcc py3-psutil \
      openssh-client
    pip3 install requests tabulate testinfra


- Debian (using deb packages):
    apt-get install -y --no-install-recommends ansible gcc libc6-dev \
      python3-distutils python3-pip python3-setuptools python3-wheel \
      python3-psutil
    pip3 install requests tabulate testinfra
"""

import argparse
import configparser
import logging
import os
import random
import shlex
import string
import subprocess
import sys
import traceback
import yaml

import cosmos

SETTINGS_FILE = './group_vars/all.yml'
ANSIBLE_INVENTORY = './inventory.yml'


def main():  # pylint: disable=too-many-statements,too-many-branches
    """ does everything """
    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description='Run bootstrap script'
        'to deploy Openappstack to a given node.')

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--create-droplet',
        action='store_true',
        help='Create droplet automatically')
    group.add_argument(
        '--droplet-id',
        metavar='ID',
        type=int,
        help='ID of droplet to deploy to')

    group.add_argument(
        '--use-existing-inventory',
        action='store_true',
        help='Assumes inventory.yml has already been generated')

    parser.add_argument(
        '--ssh-key-id',
        metavar='ID',
        type=int,
        default=411,
        help='ID of ssh key to deploy with (default: 411)')

    parser.add_argument(
        '--terminate',
        action='store_true',
        help='Terminate droplet after deploy (shutdown and delete)')

    parser.add_argument(
        '--verbose',
        action='store_true',
        help='Be more verbose')

    parser.add_argument(
        '--ansible-param',
        metavar=['PARAM[=VALUE]'],
        action='append',
        nargs=1,
        help=('forward ansible parameters to the ansible-playbook call '
              '(two dashes are prepended to PARAM)'))
    parser.add_argument(
        '--run-ansible',
        action='store_true',
        help='Runs the ansible bootstrap process')

    parser.add_argument(
        '--create-domain-records',
        action='store_true',
        help='Creates DNS entries for the cluster')

    parser.add_argument(
        '--write-behave-config',
        action='store_true',
        help='Writes a configuration file for behave with cluster information')

    args = parser.parse_args()
    verbose = args.verbose
    loglevel = logging.DEBUG if verbose else logging.INFO
    init_logging(log, loglevel)

    # Setup logging for cosmos module
    log_cosmos = logging.getLogger('cosmos')
    init_logging(log_cosmos, loglevel)

    if not args.use_existing_inventory:
        # Start bootstrapping
        if args.create_droplet:
            # Create droplet

            # image: 15 = Debian stretch-x64
            # ssh_keys
            # - 411: ci, ed25519
            # - 407: varac

            if "CI_PIPELINE_ID" in os.environ:
                instance_id = os.environ['CI_PIPELINE_ID']
            else:
                # Use random generated ID in case we're not running in gitlab CI
                # and there's no CI_PIPELINE_ID env var
                instance_id = ''.join(
                    random.choice(string.ascii_lowercase + string.digits)
                    for _ in range(10))

            droplet = cosmos.create_droplet(
                name='ci-' + instance_id,
                ssh_key_id=args.ssh_key_id,
                region='ams1',
                size=4096,
                disk=8,
                image=15)
            droplet_id = droplet['droplet']['id']
            log.info('Created droplet id: %s', droplet_id)
            cosmos.wait_for_state(droplet_id, 'running')
        else:
            droplet_id = args.droplet_id

        if verbose:
            cosmos.list_droplets()

        # Get droplet ip
        droplet = cosmos.get_droplet(droplet_id)
        droplet_ip = droplet['networks']['v4'][0]['ip_address']
        droplet_name = droplet['name']

        # Create inventory
        with open('../ansible/inventory.yml.example', 'r') as stream:
            inventory = yaml.safe_load(stream)
        inventory['all']['hosts'][droplet_name] = inventory['all']['hosts']['oas-dev']
        del inventory['all']['hosts']['oas-dev']

        inventory['all']['hosts'][droplet_name]['ansible_host'] = droplet_ip
        inventory['all']['hosts'][droplet_name]['ansible_ssh_extra_args'] = \
            '-o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no'
        inventory['all']['children']['cluster']['hosts'] = droplet_name
        inventory['all']['children']['master']['hosts'] = droplet_name
        inventory['all']['children']['worker']['hosts'] = droplet_name

        with open(ANSIBLE_INVENTORY, 'w') as stream:
            yaml.safe_dump(inventory, stream, default_flow_style=False)

        # Create settings
        with open('../ansible/group_vars/cluster/settings.yml.example',
                  'r') as stream:
            settings = yaml.safe_load(stream)

        settings['ip_address'] = droplet_ip
        settings['domain'] = droplet_name + '.ci.openappstack.net'
        settings['admin_email'] = "admin@{0}".format(settings['domain'])
        settings['acme_staging'] = True

        if not os.path.exists('./group_vars'):
            os.mkdir('./group_vars')
        with open(SETTINGS_FILE, 'w') as stream:
            yaml.safe_dump(settings, stream, default_flow_style=False)

        log.debug(yaml.safe_dump(inventory, default_flow_style=False))
        log.debug(yaml.safe_dump(settings, default_flow_style=False))

        # Wait for ssh
        cosmos.wait_for_ssh(droplet_ip)
    else:
        # Work with the master node from the inventory
        with open(ANSIBLE_INVENTORY, 'r') as stream:
            inventory = yaml.safe_load(stream)
        droplet_name = inventory['all']['children']['master']['hosts']
        droplet_ip = inventory['all']['hosts'][droplet_name]['ansible_host']
        log.info("Read data from inventory:\n\tname: %s\n\tip: %s",
                 droplet_name, droplet_ip)

        # For if write_behave_config is called later:
        settings = None

    if args.create_domain_records:
        # Create domain records
        domain_record = cosmos.create_domain_record(
            domain='openappstack.net', name=droplet_name + '.ci', data=droplet_ip, record_type='A',
            update=True)
        log.info("Domain record: %s", domain_record)

        domain_record = cosmos.create_domain_record(
            domain='openappstack.net', name='*.' + droplet_name + '.ci', data=droplet_name + '.ci',
            record_type='CNAME', update=True)
        log.info("Domain record: %s", domain_record)

        if verbose:
            cosmos.list_domain_records('openappstack.net')


    if args.run_ansible:

        run_ansible('./bootstrap.yml', args.ansible_param)

    if args.write_behave_config:
        write_behave_config(settings=settings)

    if args.terminate:
        cosmos.terminate_droplets_by_name(droplet_name)

def run_ansible(playbook, ansible_params):
    """ Calls `ansible-playbook` directly to run the specified playbook. """
    # playbook path here is relative to private_data_dir/project, see
    # https://ansible-runner.readthedocs.io/en/latest/intro.html#inputdir
    ansible_playbook_cmd = 'ansible-playbook %s' % playbook

    if ansible_params:
        for param in ansible_params:
            if len(param) > 1:
                log.warning('More than 1 parameter. Ignoring the rest! Use '
                            '--ansible-param several times to supply '
                            'more than 1 parameter')
            param = param[0]
            ansible_playbook_cmd += ' --' + param

    log.info('Running %s', ansible_playbook_cmd)

    result = subprocess.run(shlex.split(ansible_playbook_cmd))

    if result.returncode > 0:
        try:
            raise RuntimeError('Playbook failed with rc %s.'
                               % result.returncode)
        except RuntimeError:
            traceback.print_exc()
            sys.exit(result.returncode)

def write_behave_config(settings=None):
    """ Write behave config file for later use """
    if settings is None:
        with open(SETTINGS_FILE) as stream:
            settings = yaml.safe_load(stream)
    behave_config = configparser.ConfigParser()
    behave_config['behave'] = {'stop': True}
    behave_config['behave.userdata'] = {}
    with open('./behave/behave.ini', 'w') as configfile:
        behave_config.write(configfile)

def init_logging(logger, loglevel):
    """
    Configure logging.

    - debug and info go to stdout
    - warning and above go to stderr
    """
    logger.setLevel(loglevel)
    stdout = logging.StreamHandler(sys.stdout)
    stdout.setLevel(loglevel)
    stdout.addFilter(lambda record: record.levelno <= logging.INFO)

    stderr = logging.StreamHandler()
    stderr.setLevel(logging.WARNING)

    logger.addHandler(stdout)
    logger.addHandler(stderr)

if __name__ == "__main__":
    # Setup logging for this script
    log = logging.getLogger(__name__)  # pylint: disable=invalid-name

    main()
